%% ``The contents of this file are subject to the Erlang Public License,
%% Version 1.1, (the "License"); you may not use this file except in
%% compliance with the License. You should have received a copy of the
%% Erlang Public License along with this software. If not, it can be
%% retrieved via the world wide web at http://www.erlang.org/.
%% 
%% Software distributed under the License is distributed on an "AS IS"
%% basis, WITHOUT WARRANTY OF ANY KIND, either express or implied. See
%% the License for the specific language governing rights and limitations
%% under the License.
%% 
%% The Initial Developer of the Original Code is Ericsson Utvecklings AB.
%% Portions created by Ericsson are Copyright 1999, Ericsson Utvecklings
%% AB. All Rights Reserved.''
%% 
%%     $Id$
%%
-module(port_SUITE).

%%%
%%% Author: Bjorn Gustavsson; iter_max_ports contributed by Peter Hogfeldt.
%%%

%%
%% There are a lot of things to test with open_port(Name, Settings).
%%
%%   Name can be
%%
%%       {spawn, Command}
%%         which according to The Book and the manual page starts an
%%         external program. That is not true. It might very well be
%%         a linked-in program (the notion of 'linked-in driver' is
%%         silly, since any driver is 'linked-in').
%%	  [Spawn of external program is tested.]
%%
%%       Atom
%%         Read all contents of Atom, or write to it.
%%
%%       {fd, In, Out}
%%       Open file descriptors In and Out. [Not tested]
%%
%%   PortSettings can be
%%
%%       {packet, N}
%%         N is 1, 2 or 4.
%%
%%       stream (default)
%%         Without packet length.
%%
%%       use_stdio (default for spawned ports)
%%         The spawned process use file descriptors 0 and 1 for I/O.
%%
%%       nouse_stdio					[Not tested]
%%         Use filedescriptors 3 and 4.  This option is probably only
%%	   meaningful on Unix.
%%
%%       in (default for Atom)
%%         Input only (from Erlang's point of view).
%%
%%       out
%%         Output only (from Erlang's point of view).
%%
%%       binary
%%         The port is a binary port, i.e. messages received and sent
%%         to a port are binaries.
%%
%%       eof
%%         Port is not closed on eof and will not send an exit signal,
%%         instead it will send a {Port, eof} to the controlling process
%%         (output can still be sent to the port (??)).
%%


-export([all/1, stream/1, stream_small/1, stream_big/1,
	 basic_ping/1, slow_writes/1, bad_packet/1,
	 multiple_packets/1, mul_basic/1, mul_slow_writes/1,
	 dying_port/1, port_program_with_path/1,
	 open_input_file_port/1, open_output_file_port/1,
	 iter_max_ports/1, eof/1, input_only/1, output_only/1,
	 name1/1,
	 t_binary/1, options/1, parallell/1, t_exit/1,
	 env/1, cd/1, exit_status/1,
	 tps/1, tps_16_bytes/1, tps_1K/1, line/1, stderr_to_stdout/1,
	 otp_3906/1, otp_4389/1]).

%% Internal exports.
-export([tps/3]).
-export([otp_3906_forker/5, otp_3906_start_forker_starter/4]).

-include("test_server.hrl").
-include_lib("kernel/include/file.hrl").

all(suite) -> [stream, basic_ping, slow_writes, bad_packet,
	       options, multiple_packets, parallell,
	       dying_port, port_program_with_path,
	       open_input_file_port, open_output_file_port,
	       name1,
	       env, cd, exit_status,
	       iter_max_ports, t_exit, tps, line, stderr_to_stdout,
	       otp_3906, otp_4389].

stream(suite) -> [stream_small, stream_big].

%% Test that we can send a stream of bytes and get it back.
%% We will send only a small amount of data, to avoid deadlock.

stream_small(suite) -> [];
stream_small(Config) when list(Config) ->
    ?line Dog = test_server:timetrap(test_server:seconds(10)),
    ?line stream_ping(Config, 512, "", []),
    ?line stream_ping(Config, 1777, "", []),
    ?line stream_ping(Config, 1777, "-s512", []),
    ?line test_server:timetrap_cancel(Dog),
    ok.

%% Send big amounts of data (much bigger than the buffer size in port test).
%% This will deadlock the emulator if the spawn driver haven't proper
%% non-blocking reads and writes.

stream_big(suite) -> [];
stream_big(Config) when list(Config) ->
    ?line Dog = test_server:timetrap(test_server:seconds(180)),
    case os:type() of 
	vxworks ->
	    %% Don't stress VxWorks too much
	    ?line stream_ping(Config, 43755, "", []),
	    ?line stream_ping(Config, 51255, "", []),
	    ?line stream_ping(Config, 52345, " -s40000", []);
	_ ->
	    ?line stream_ping(Config, 43755, "", []),
	    ?line stream_ping(Config, 100000, "", []),
	    ?line stream_ping(Config, 77777, " -s40000", [])
    end,
    ?line test_server:timetrap_cancel(Dog),
    ok.

%% Sends packet with header size of 1, 2, and 4, with packets of various
%% sizes.

basic_ping(suite) -> [];
basic_ping(Config) when list(Config) ->
    ?line Dog = test_server:timetrap(test_server:seconds(120)),
    ?line ping(Config, sizes(1), 1, "", []),
    ?line ping(Config, sizes(2), 2, "", []),
    ?line ping(Config, sizes(4), 4, "", []),
    ?line test_server:timetrap_cancel(Dog),
    ok.

%% Let the port program insert delays between characters sent back to
%% Erlang, to test that the Erlang emulator can handle a packet coming in
%% small chunks rather than all at once.

slow_writes(suite) -> [];
slow_writes(Config) when list(Config) ->
    ?line Dog = test_server:timetrap(test_server:seconds(20)),
    ?line ping(Config, [8], 4, "-s1", []),
    ?line ping(Config, [10], 2, "-s2", []),
    ?line test_server:timetrap_cancel(Dog),
    ok.

bad_packet(doc) ->
    ["Test that we get {'EXIT', Port, einval} if we try to send a bigger "
     "packet than the packet header allows."];
bad_packet(suite) -> [];
bad_packet(Config) when list(Config) ->
    ?line Dog = test_server:timetrap(test_server:seconds(10)),
    ?line PortTest = port_test(Config),
    ?line process_flag(trap_exit, true),

    ?line bad_packet(PortTest, 1, 256),
    ?line bad_packet(PortTest, 1, 257),
    ?line bad_packet(PortTest, 2, 65536),
    ?line bad_packet(PortTest, 2, 65537),

    ?line test_server:timetrap_cancel(Dog),
    ok.

bad_packet(PortTest, HeaderSize, PacketSize) ->
    %% Intentionally no ?line macros.
    P = open_port({spawn, PortTest}, [{packet, HeaderSize}]),
    P ! {self(), {command, make_zero_packet(PacketSize)}},
    receive
	{'EXIT', P, einval} -> ok;
	Other -> test_server:fail({unexpected_message, Other})
    end.

make_zero_packet(0) -> [];
make_zero_packet(N) when N rem 2 == 0 ->
    P = make_zero_packet(N div 2),
    [P|P];
make_zero_packet(N) ->
    P = make_zero_packet(N div 2),
    [0, P|P].
    
%% Tests various options (stream and {packet, Number} are implicitly
%% tested in other test cases).

options(suite) -> [t_binary, eof, input_only, output_only].

%% Tests the 'binary' option for a port.

t_binary(suite) -> [];
t_binary(Config) when list(Config) ->
    ?line Dog = test_server:timetrap(test_server:seconds(300)),

    %% Packet mode.
    ?line ping(Config, sizes(1), 1, "", [binary]),
    ?line ping(Config, sizes(2), 2, "", [binary]),
    ?line ping(Config, sizes(4), 4, "", [binary]),

    %% Stream mode.
    case os:type() of
	vxworks ->
	    %% don't stress VxWorks too much
	    ?line stream_ping(Config, 435, "", [binary]),
	    ?line stream_ping(Config, 43755, "", [binary]),
	    ?line stream_ping(Config, 50000, "", [binary]);
	_ ->
	    ?line stream_ping(Config, 435, "", [binary]),
	    ?line stream_ping(Config, 43755, "", [binary]),
	    ?line stream_ping(Config, 100000, "", [binary])
    end,

    ?line test_server:timetrap_cancel(Dog),
    ok.

name1(suite) -> [];
name1(Config) when list(Config) ->
    ?line Dog = test_server:timetrap(test_server:seconds(100)),
    ?line PortTest = port_test(Config),
    ?line Command = lists:concat([PortTest, " "]),
    ?line P = open_port({spawn, Command}, []),
    ?line register(myport, P),
    ?line P = whereis(myport),
    Text = "hej",
    ?line myport ! {self(), {command, Text}},
    ?line receive
	      {P, {data, Text}} ->
		  ok
	  end,
    ?line myport ! {self(), close},
    ?line receive
	      {P, closed} -> ok
	  end,
    ?line undefined = whereis(myport),
    ?line test_server:timetrap_cancel(Dog),
    ok.

%% Test that the 'eof' option works.

eof(suite) -> [];
eof(Config) when list(Config) ->
    ?line Dog = test_server:timetrap(test_server:seconds(100)),
    ?line PortTest = port_test(Config),
    ?line Command = lists:concat([PortTest, " -h0 -q"]),
    ?line P = open_port({spawn, Command}, [eof]),
    ?line receive
	      {P, eof} ->
		  ok
	  end,
    ?line P ! {self(), close},
    ?line receive
	      {P, closed} -> ok
	  end,
    ?line test_server:timetrap_cancel(Dog),
    ok.

%% Tests that the 'in' option for a port works.

input_only(suite) -> [];
input_only(Config) when list(Config) ->
    ?line Dog = test_server:timetrap(test_server:seconds(300)),
    ?line expect_input(Config, [0, 1, 10, 13, 127, 128, 255], 1, "", [in]),
    ?line expect_input(Config, [0, 1, 255, 2048], 2, "", [in]),
    ?line expect_input(Config, [0, 1, 255, 2048], 4, "", [in]),
    ?line expect_input(Config, [0, 1, 10, 13, 127, 128, 255],
		       1, "", [in, binary]),
    ?line test_server:timetrap_cancel(Dog),
    ok.

%% Tests that the 'out' option for a port works.

output_only(suite) -> [];
output_only(Config) when list(Config) ->
    ?line Dog = test_server:timetrap(test_server:seconds(100)),
    ?line Dir = ?config(priv_dir, Config),
    ?line Filename = filename:join(Dir, "output_only_stream"),
    ?line output_and_verify(Config, Filename, "-h0",
			    random_packet(35777, "echo")),
    ?line test_server:timetrap_cancel(Dog),
    ok.

output_and_verify(Config, Filename, Options, Data) ->
    ?line PortTest = port_test(Config),
    ?line Command = lists:concat([PortTest, " ",
				  Options, " -o", Filename]),
    ?line Port = open_port({spawn, Command}, [out]),
    ?line Port ! {self(), {command, Data}},
    ?line Port ! {self(), close},
    ?line receive
	      {Port, closed} -> ok
	  end,
    Wait_time = case os:type() of
		    vxworks -> 5000;
		    _ -> 500
		end,
    ?line test_server:sleep(Wait_time),
    ?line {ok, Written} = file:read_file(Filename),
    ?line Data = binary_to_list(Written),
    ok.

%% Test that receiving several packages written in the same
%% write operation works.

multiple_packets(suite) -> [mul_basic, mul_slow_writes].

mul_basic(suite) -> [];
mul_basic(doc) -> "Basic test of receiving multiple packages, written in one operation by the other end.";
mul_basic(Config) when list(Config) ->
    ?line Dog = test_server:timetrap(test_server:seconds(600)),
    case os:type() of
	vxworks ->
	    %% don't stress vxworks too much
	    ?line expect_input(Config, [0, 1, 255, 10, 13], 1, "", []),
	    ?line expect_input(Config, [0, 10, 13, 1600, 8191, 16383], 2, "", []),
	    ?line expect_input(Config, [10, 35000], 4, "", []);
	_ ->
	    ?line expect_input(Config, [0, 1, 255, 10, 13], 1, "", []),
	    ?line expect_input(Config, [0, 10, 13, 1600, 32767, 65535], 2, "", []),
	    ?line expect_input(Config, [10, 70000], 4, "", [])
    end,
    ?line test_server:timetrap_cancel(Dog),
    ok.

%% Test reading a buffer consisting of several packets, some
%% of which might be incomplete.  (The port program builds
%% a buffer with several packets, but writes it in chunks with
%% delays in between.)

mul_slow_writes(suite) -> [];
mul_slow_writes(Config) when list(Config) ->
    ?line Dog = test_server:timetrap(test_server:seconds(250)),
    ?line expect_input(Config, [0, 20, 255, 10, 1], 1, "-s64", []),
    ?line test_server:timetrap_cancel(Dog),
    ok.

%% Runs several port tests in parallell.  Each individual test
%% finishes in about 5 seconds.  Running in parallell, all tests
%% should also finish in about 5 seconds.

parallell(suite) -> [];
parallell(Config) when list(Config) ->
    ?line Dog = test_server:timetrap(test_server:seconds(300)),
    ?line Testers =
	[fun() -> stream_ping(Config, 1007, "-s100", []) end,
	 fun() -> stream_ping(Config, 10007, "-s1000", []) end,
	 fun() -> stream_ping(Config, 10007, "-s1000", []) end,

	 fun() -> expect_input(Config, [21, 22, 23, 24, 25], 1,
			       "-s10", [in]) end,

	 fun() -> ping(Config, [10], 1, "-d", []) end,
	 fun() -> ping(Config, [20000], 2, "-d", []) end,
	 fun() -> ping(Config, [101], 1, "-s10", []) end,
	 fun() -> ping(Config, [1001], 2, "-s100", []) end,
	 fun() -> ping(Config, [10001], 4, "-s1000", []) end,

	 fun() -> ping(Config, [501, 501], 2, "-s100", []) end,
	 fun() -> ping(Config, [11, 12, 13, 14, 11], 1, "-s5", []) end],
    ?line process_flag(trap_exit, true),
    ?line Pids = lists:map(fun fun_spawn/1, Testers),
    ?line wait_for(Pids),
    ?line test_server:timetrap_cancel(Dog),
    ok.

wait_for([]) ->
    ok;
wait_for(Pids) ->
    io:format("Waiting for ~p", [Pids]),
    receive
	{'EXIT', Pid, normal} ->
	    wait_for(lists:delete(Pid, Pids));
	Other ->
	    test_server:fail({bad_exit, Other})
    end.

%% Tests starting port programs that terminate by themselves.
%% This used to cause problems on Windows.

dying_port(suite) -> [];
dying_port(Config) when list(Config) ->
    ?line Dog = test_server:timetrap(test_server:seconds(150)),
    ?line process_flag(trap_exit, true),

    ?line P1 = make_dying_port(Config),
    ?line P2 = make_dying_port(Config),
    ?line P3 = make_dying_port(Config),
    ?line P4 = make_dying_port(Config),
    ?line P5 = make_dying_port(Config),

    %% This should be big enough to be sure to block in the write.
    ?line Garbage = random_packet(16384),

    ?line P1 ! {self(), {command, Garbage}},
    ?line P3 ! {self(), {command, Garbage}},
    ?line P5 ! {self(), {command, Garbage}},

    ?line wait_for_port_exit(P1),
    ?line wait_for_port_exit(P2),
    ?line wait_for_port_exit(P3),
    ?line wait_for_port_exit(P4),
    ?line wait_for_port_exit(P5),

    ?line test_server:timetrap_cancel(Dog),
    ok.

wait_for_port_exit(Port) ->
    receive
	{'EXIT', Port, _} ->
	    ok
    end.

make_dying_port(Config) when list(Config) ->
    PortTest = port_test(Config),
    Command = lists:concat([PortTest, " -h0 -d -q"]),
    open_port({spawn, Command}, [stream]).

%% Tests that port program with complete path (but without any
%% .exe extension) can be started, even if there is a file with
%% the same name but without the extension in the same directory.
%% (In practice, the file with the same name could be a Unix
%% executable.)
%%
%% This used to failed on Windows (the .exe extension had to be
%% explicitly given).
%%
%% This testcase works on Unix, but is not very useful.

port_program_with_path(suite) -> [];
port_program_with_path(Config) when list(Config) ->
    ?line Dog = test_server:timetrap(test_server:seconds(100)),
    ?line DataDir = ?config(data_dir, Config),
    ?line PrivDir = ?config(priv_dir, Config),
    
    %% Create a copy of the port test program in a directory not
    %% included in PATH (i.e. in priv_dir), with the name 'my_port_test.exe'.
    %% Also, place a file named 'my_port_test' in the same directory.
    %% This used to confuse the CreateProcess() call in spawn driver.
    %% (On Unix, there will be a single file created, which will be
    %% a copy of the port program.)
    
    ?line PortTest = os:find_executable("port_test", DataDir),
    io:format("os:find_executable(~p, ~p) returned ~p",
	      ["port_test", DataDir, PortTest]),
    ?line {ok, PortTestPgm} = file:read_file(PortTest),
    ?line NewName = filename:join(PrivDir, filename:basename(PortTest)),
    ?line RedHerring = filename:rootname(NewName),
    ?line ok = file:write_file(RedHerring, "I'm just here to confuse.\n"),
    ?line ok = file:write_file(NewName, PortTestPgm),
    ?line ok = file:write_file_info(NewName, #file_info{mode=8#111}),
    ?line PgmWithPathAndNoExt = filename:rootname(NewName),
    
    %% Open the port using the path to the copied port test program,
    %% but without the .exe extension, and verified that it was started.
    %%
    %% If the bug is present the open_port call will fail with badarg.
    
    ?line Command = lists:concat([PgmWithPathAndNoExt, " -h2"]),
    %% allow VxWorks time to write file
    case os:type() of	
	vxworks -> test_server:sleep(2500);
	_ -> time
    end,
    ?line P = open_port({spawn, Command}, [{packet, 2}]),
    ?line Message = "echo back to me",
    ?line P ! {self(), {command, Message}},
    ?line receive
	      {P, {data, Message}} ->
		  ok
	  end,
    ?line test_server:timetrap_cancel(Dog),
    ok.


%% Tests that files can be read using open_port(Filename, [in]).
%% This used to fail on Windows.
open_input_file_port(suite) -> [];
open_input_file_port(Config) when list(Config) ->
    ?line Dog = test_server:timetrap(test_server:seconds(10)),
    ?line PrivDir = ?config(priv_dir, Config),
    
    %% Create a file with the file driver and read it back using
    %% open_port/2.

    ?line MyFile1 = filename:join(PrivDir, "my_input_file"),
    ?line FileData1 = "An input file",
    ?line ok = file:write_file(MyFile1, FileData1),
    case os:type() of
	vxworks ->
	    %% Can't open input file with vanilla driver on VxWorks
	    ?line Oldexit = process_flag(trap_exit, true),
	    ?line case catch open_port(MyFile1, [in]) of
		      {'EXIT', {badarg, _}} ->
			  ok
		  end;
	_ ->
	    ?line case open_port(MyFile1, [in]) of
		      InputPort when port(InputPort) ->
			  ?line receive
				    {InputPort, {data, FileData1}} ->
					ok
				end
		  end
    end,
    ?line test_server:timetrap_cancel(Dog),
    ok.

%% Tests that files can be written using open_port(Filename, [out]).
open_output_file_port(suite) -> [];
open_output_file_port(Config) when list(Config) ->
    ?line Dog = test_server:timetrap(test_server:seconds(100)),
    ?line PrivDir = ?config(priv_dir, Config),

    %% Create a file with open_port/2 and read it back with
    %% the file driver.

    ?line MyFile2 = filename:join(PrivDir, "my_output_file"),
    ?line FileData2_0 = "A file created ",
    ?line FileData2_1 = "with open_port/2.\n",
    ?line FileData2 = FileData2_0 ++ FileData2_1,
    ?line OutputPort = open_port(MyFile2, [out]),
    ?line OutputPort ! {self(), {command, FileData2_0}},
    ?line OutputPort ! {self(), {command, FileData2_1}},
    ?line OutputPort ! {self(), close},
    ?line {ok, Bin} = file:read_file(MyFile2),
    ?line FileData2 = binary_to_list(Bin),

    ?line test_server:timetrap_cancel(Dog),
    ok.

%%
%% Open as many ports as possible. Do this several times and check
%% that we get the same number of ports every time.
%%

iter_max_ports(suite) -> [];
iter_max_ports(Config) when list(Config) ->
    ?line Dog = test_server:timetrap(test_server:minutes(20)),
    ?line PortTest = port_test(Config),
    ?line Command = lists:concat([PortTest, " -h0 -q"]),
    ?line Iters = case os:type() of
		      {win32,_} -> 4;
		      _ -> 10
		  end,
    ?line L = do_iter_max_ports(Iters, Command),
    io:format("Result: ~p",[L]),
    ?line all_equal(L),
    ?line test_server:timetrap_cancel(Dog),
    {comment, "Max ports: " ++ integer_to_list(hd(L))}.

do_iter_max_ports(N, Command) when N > 0 ->
    [max_ports(Command)| do_iter_max_ports(N-1, Command)];
do_iter_max_ports(_, _) ->
    [].

all_equal([E, E| T]) ->
    all_equal([E| T]);
all_equal([E]) ->
    ok;
all_equal([]) ->
    ok.

max_ports(Command) ->
    test_server:sleep(500),
    ?line Ps = open_ports({spawn, Command}, [eof]),
    ?line N = length(Ps),
    ?line close_ports(Ps),
    io:format("Got ~p ports\n",[N]),
    N.

close_ports([P|Ps]) ->
    P ! {self(), close},
    receive
	{P,closed} ->
	    ok
    end,
    close_ports(Ps);
close_ports([]) ->
    ok.

open_ports(Name, Settings) ->
    test_server:sleep(50),
    case catch open_port(Name, Settings) of
	P when port(P) ->
	    [P| open_ports(Name, Settings)];
	{'EXIT', {Code, _}} ->
	    case Code of
		enfile ->
		    [];
		emfile ->
		    [];
		Other ->
		    ?line test_server:fail({open_ports, Other})
	    end;
	Other ->
	    ?line test_server:fail({open_ports, Other})
    end.

%% Tests that exit(Port, Term) works (has been known to crash the emulator).

t_exit(suite) -> [];
t_exit(Config) when list(Config) ->
    ?line process_flag(trap_exit, true),
    ?line Pid = fun_spawn(fun suicide_port/1, [Config]),
    ?line receive
	      {'EXIT', Pid, die} ->
		  ok;
	      Other ->
		  test_server:fail({bad_message, Other})
	  end.

suicide_port(Config) when list(Config) ->
    ?line Port = port_expect(Config, [], 0, "", []),
    ?line exit(Port, die).

tps(suite) -> [tps_16_bytes, tps_1K].

tps_16_bytes(doc) -> "";
tps_16_bytes(suite) -> [];
tps_16_bytes(Config) when list(Config) ->
    ?line tps(16, Config).

tps_1K(doc) -> "";
tps_1K(suite) -> [];
tps_1K(Config) when list(Config) ->
    ?line tps(1024, Config).

tps(Size, Config) ->
    ?line Dog = test_server:timetrap(test_server:seconds(300)),
    ?line PortTest = port_test(Config),
    ?line Cmd = lists:concat([PortTest, " -h2"]),
    ?line Packet = list_to_binary(random_packet(Size, "e")),
    ?line Port = open_port({spawn, PortTest}, [binary, {packet, 2}]),
    ?line Transactions = 10000,
    ?line {Elapsed, ok} = test_server:timecall(?MODULE, tps,
					       [Port, Packet, Transactions]),
    ?line test_server:timetrap_cancel(Dog),
    {comment, integer_to_list(trunc(Transactions/Elapsed+0.5)) ++ " transactions/s"}.

tps(Port, Packet, 0) ->
    ok;
tps(Port, Packet, N) ->
    ?line port_command(Port, Packet),
    ?line receive
	      {Port, {data, Packet}} ->
		  ?line tps(Port, Packet, N-1);
	      Other ->
		  ?line test_server:fail({bad_message, Other})
	  end.

%% Line I/O test
line(suite)->
    [];
line(Config) when list(Config) ->
    ?line Siz = 110,
    ?line Dog = test_server:timetrap(test_server:seconds(300)),
    ?line Packet1 = random_packet(Siz),
    ?line Packet2 = random_packet(Siz div 2),
    %% Test that packets are split into lines
    ?line port_expect(Config,[{lists:append([Packet1, io_lib:nl(), Packet2,
					     io_lib:nl()]),
			       [{eol, Packet1}, {eol, Packet2}]}],
		      0, "", [{line,Siz}]),
    %% Test the same for binaries
    ?line port_expect(Config,[{lists:append([Packet1, io_lib:nl(), Packet2,
					     io_lib:nl()]),
			       [{eol, Packet1}, {eol, Packet2}]}],
		      0, "", [{line,Siz},binary]),
    %% Test that too long lines get split
    ?line port_expect(Config,[{lists:append([Packet1, io_lib:nl(), Packet1,
					     Packet2, io_lib:nl()]),
			       [{eol, Packet1}, {noeol, Packet1},
				{eol, Packet2}]}], 0, "", [{line,Siz}]),
    %% Test that last output from closing port program gets received.
    ?line L1 = lists:append([Packet1, io_lib:nl(), Packet2]),
    ?line S1 = lists:flatten(io_lib:format("-l~w", [length(L1)])),
    io:format("S1 = ~w, L1 = ~w~n", [S1,L1]),
    ?line port_expect(Config,[{L1,
			       [{eol, Packet1}, {noeol, Packet2}, eof]}], 0, 
		      S1, [{line,Siz},eof]),
    %% Test that lonely <CR> Don't get treated as newlines
    ?line port_expect(Config,[{lists:append([Packet1, [13], Packet2,
					     io_lib:nl()]),
			       [{noeol, Packet1}, {eol, [13 |Packet2]}]}],
		      0, "", [{line,Siz}]),
    %% Test that packets get built up to lines (delayed output from
    %% port program)
    ?line port_expect(Config,[{Packet2,[]},
			      {lists:append([Packet2, io_lib:nl(),
					     Packet1, io_lib:nl()]),
			       [{eol, lists:append(Packet2, Packet2)},
				{eol, Packet1}]}], 0, "-d", [{line,Siz}]),
    %% Test that we get badarg if trying both packet and line
    ?line bad_argument(Config, [{packet, 5}, {line, 5}]),
    ?line test_server:timetrap_cancel(Dog),
    ok.

%%% Redirection of stderr test
stderr_to_stdout(suite) ->
    [];
stderr_to_stdout(doc) ->
    "Test that redirection of standard error to standard output works.";
stderr_to_stdout(Config) when list(Config) ->
    ?line Dog = test_server:timetrap(test_server:seconds(60)),
    %% See that it works
    ?line Packet = random_packet(10),
    ?line port_expect(Config,[{Packet,[Packet]}], 0, "-e -l10",
		      [stderr_to_stdout]),
    %% ?line stream_ping(Config, 10, "-e", [stderr_to_stdout]),
    %% See that it doesn't always happen (will generate garbage on stderr)
    ?line port_expect(Config,[{Packet,[eof]}], 0, "-e -l10", [line,eof]),
    ?line test_server:timetrap_cancel(Dog),
    ok.


bad_argument(Config, ArgList) ->
    PortTest = port_test(Config),
    case catch open_port({spawn, PortTest}, ArgList) of
	{'EXIT', {badarg, _}} ->
	    ok
    end.
    

%% 'env' option
%% (Can perhaps be made smaller by calling the other utility functions
%% in this module.)
env(suite) ->
    [];
env(doc) ->
    ["Test that the 'env' option works"];
env(Config)  when list(Config) ->
    case os:type() of
	vxworks ->
	    {skipped,"Environments not implemented on VxWorks (could be...)"};
	_ ->
	    env2(Config)
    end.

env2(Config) ->
    ?line Dog = test_server:timetrap(test_server:seconds(60)),

    ?line Program = atom_to_list(lib:progname()),
    ?line Cmd = Program ++ " -pz " ++ ?config(data_dir, Config) ++
	" -noshell -s port_test env plupp -s erlang halt",
    ?line _ = open_port({spawn, Cmd},
			[{env, [{"plupp", "apa"}]},
			 {line, 80}]),
    ?line receive
	      {_, {data, {eol,"1apa"}}} ->
		  ok;
	      Other1 ->
		  ?line test_server:fail({env, Other1})
	  end,

    ?line _ = open_port({spawn, Cmd},
			[{env, [{"plupp", false}]},
			 {line, 80}]),
    ?line receive
	      {_, {data, {eol, "0"}}} ->
		  ok;
	      Other2 ->
		  ?line test_server:fail({env, Other2})
	  end,

    ?line test_server:timetrap_cancel(Dog),
    ok.

%% 'cd' option
%% (Can perhaps be made smaller by calling the other utility functions
%% in this module.)
cd(suite) ->
    [];
cd(doc) ->
    ["Test that the 'cd' option works"];
cd(Config)  when list(Config) ->
    case os:type() of
	vxworks ->
	    {skipped,"Task specific directories does not exist on VxWorks"};
	_ ->
	    cd2(Config)
    end.
cd2(Config) ->
    ?line Dog = test_server:timetrap(test_server:seconds(60)),

    ?line Program = atom_to_list(lib:progname()),
    ?line DataDir = ?config(data_dir, Config),
    ?line TestDir = filename:join(DataDir, "dir"),
    ?line Cmd = Program ++ " -pz " ++ DataDir ++
	" -noshell -s port_test pwd -s erlang halt",
    ?line _ = open_port({spawn, Cmd},
			[{cd, TestDir},
			 {line, 256}]),
    ?line receive
	      {_, {data, {eol, String}}} ->
		  case filename_equal(String, TestDir) of
		      true ->
			  ok;
		      false ->
			  ?line test_server:fail({cd, String})
		  end;
	      Other2 ->
		  ?line test_server:fail({env, Other2})
	  end,

    ?line test_server:timetrap_cancel(Dog),
    ok.

filename_equal(A, B) ->
    case os:type() of
	{win32, _} ->
	    win_filename_equal(A, B);
	_ ->
	    A == B
    end.

win_filename_equal([], []) ->
    true;
win_filename_equal([], _) ->
    false;
win_filename_equal(_, []) ->
    false;
win_filename_equal([C1 | Rest1], [C2 | Rest2]) ->
    case tolower(C1) == tolower(C2) of
	true ->
	    win_filename_equal(Rest1, Rest2);
	false ->
	    false
    end.

tolower(C) when C >= $A, C =< $Z ->
    C + 32;
tolower(C) ->
    C.

otp_3906(suite) ->
    [];
otp_3906(doc) ->
    ["Tests that child process deaths are managed correctly when there are "
     " a large amount of concurrently dying children. See ticket OTP-3906."];
otp_3906(Config)  when list(Config) ->
    case os:type() of
	{unix, OSName} ->
	    otp_3906(Config, OSName);
	_ ->
	    {skipped, "Only run on Unix systems"}
    end.

-define(OTP_3906_CHILDREN,     1000).
-define(OTP_3906_EXIT_STATUS,  17).
-define(OTP_3906_PROGNAME,     "otp_3906").
-define(OTP_3906_TICK_TIMEOUT, 5000).
-define(OTP_3906_OSP_P_ERLP,   10).
-define(OTP_3906_MAX_CONC_OSP, 50).

otp_3906(Config, OSName) ->
    ?line TSDir = filename:dirname(code:which(test_server)),
    ?line {ok, Variables} = file:consult(filename:join(TSDir, "variables")),
    case lists:keysearch('CC', 1, Variables) of
	{value,{'CC', CC}} ->
	    SuiteDir = filename:dirname(code:which(?MODULE)),
	    PrivDir = ?config(priv_dir, Config),
	    Prog = otp_3906_make_prog(CC, PrivDir),
	    {ok, Node} = test_server:start_node(otp_3906,
						slave,
						[{args, " -pa " ++ SuiteDir},
						 {linked, false}]),
	    OP = process_flag(priority, max),
	    OTE = process_flag(trap_exit, true),
	    FS = spawn_link(Node,
			    ?MODULE,
			    otp_3906_start_forker_starter,
			    [?OTP_3906_CHILDREN, [], self(), Prog]),
	    Result = receive
			 {'EXIT', ForkerStarter, Reason} ->
			     {failed, Reason};
			 {emulator_pid, EmPid} ->
			     case otp_3906_wait_result(FS, 0, 0) of
				 {succeded,
				  ?OTP_3906_CHILDREN,
				  ?OTP_3906_CHILDREN} ->
				     succeded;
				 {succeded, Forked, Exited} ->
				     otp_3906_list_defunct(EmPid, OSName),
				     {failed,
				      {mismatch,
				       {forked, Forked},
				       {exited, Exited}}};
				 Res ->
				     otp_3906_list_defunct(EmPid, OSName),
				     Res
			     end
		     end,
	    process_flag(trap_exit, OTE),
	    process_flag(priority, OP),
	    test_server:stop_node(Node),
	    case Result of
		succeded ->
		    ok;
		_ ->
		    ?line test_server:fail(Result)
	    end;
	_ ->
	    {skipped, "No C compiler found"}
    end.

otp_3906_list_defunct(EmPid, OSName) ->
    % Guess ps switches to use and what to grep for (could be improved)
    {Switches, Zombie} = case OSName of
			     freebsd ->
				 {"-ajx", "Z"};
			     _ ->
				 {"-ef", "[dD]efunct"}
			 end,
    test_server:format("Emulator pid: ~s~n"
		       "Listing of zombie processes:~n"
		       "~s~n",
		       [EmPid,
			otp_3906_htmlize(os:cmd("ps "
						++ Switches
						++ " | grep "
						++ Zombie))]).

otp_3906_htmlize([]) ->
    [];
otp_3906_htmlize([C | Cs]) ->
    case [C] of
	"<" -> "&lt;" ++ otp_3906_htmlize(Cs);
	">" -> "&gt;" ++ otp_3906_htmlize(Cs);
	_ ->   [C | otp_3906_htmlize(Cs)]
    end.

otp_3906_make_prog(CC, PrivDir) ->
    SrcFileName = filename:join(PrivDir, ?OTP_3906_PROGNAME ++ ".c"),
    TrgtFileName = filename:join(PrivDir, ?OTP_3906_PROGNAME),
    {ok, SrcFile} = file:open(SrcFileName, write),
    io:format(SrcFile,
	      "int           ~n"
	      "main(void)    ~n"
	      "{             ~n"
	      "   return ~p; ~n"
	      "}             ~n",
	      [?OTP_3906_EXIT_STATUS]),
    file:close(SrcFile),
    os:cmd(CC ++ " " ++ SrcFileName ++ " -o " ++ TrgtFileName),
    TrgtFileName.


otp_3906_wait_result(ForkerStarter, F, E) ->
    receive
	{'EXIT', ForkerStarter, Reason} ->
	    {failed, {Reason, {forked, F}, {exited, E}}};
	forked ->
	    otp_3906_wait_result(ForkerStarter, F+1, E);
	exited ->
	    otp_3906_wait_result(ForkerStarter, F, E+1);
	tick ->
	    otp_3906_wait_result(ForkerStarter, F, E);
	succeded ->
	    {succeded, F, E}
    after
	?OTP_3906_TICK_TIMEOUT ->
	    unlink(ForkerStarter),
	    exit(ForkerStarter, timeout),
	    {failed, {timeout, {forked, F}, {exited, E}}}
    end.

otp_3906_collect([], _) ->
    done;
otp_3906_collect(RefList, Sup) ->
    otp_3906_collect(otp_3906_collect_one(RefList, Sup), Sup).

otp_3906_collect_one(RefList, Sup) ->
    receive
	Ref when reference(Ref) ->
	    Sup ! tick,
	    lists:delete(Ref, RefList)
    end.
    
otp_3906_start_forker(N, Sup, Prog) ->
    Ref = make_ref(),
    erlang:spawn_opt(?MODULE,
		     otp_3906_forker,
		     [N, self(), Ref, Sup, Prog],
		     [link, {priority, max}]),
    Ref.

otp_3906_start_forker_starter(N, RefList, Sup, Prog) ->
    process_flag(priority, max),
    EmPid = os:getpid(),
    Sup ! {emulator_pid, EmPid},
    otp_3906_forker_starter(N, RefList, Sup, Prog).

otp_3906_forker_starter(0, RefList, Sup, _) ->
    otp_3906_collect(RefList, Sup),
    unlink(Sup),
    Sup ! succeded;
otp_3906_forker_starter(N, RefList, Sup, Prog)
  when length(RefList) >= ?OTP_3906_MAX_CONC_OSP ->
    otp_3906_forker_starter(N, otp_3906_collect_one(RefList, Sup), Sup, Prog);
otp_3906_forker_starter(N, RefList, Sup, Prog)
  when integer(N), N > ?OTP_3906_OSP_P_ERLP ->
    otp_3906_forker_starter(N-?OTP_3906_OSP_P_ERLP,
			    [otp_3906_start_forker(?OTP_3906_OSP_P_ERLP,
						   Sup,
						   Prog)|RefList],
			    Sup,
			    Prog);
otp_3906_forker_starter(N, RefList, Sup, Prog) when integer(N) ->
    otp_3906_forker_starter(0,
			    [otp_3906_start_forker(N,
						   Sup,
						   Prog)|RefList],
			    Sup,
			    Prog).

otp_3906_forker(0, Parent, Ref, _, _) ->
    unlink(Parent),
    Parent ! Ref;
otp_3906_forker(N, Parent, Ref, Sup, Prog) ->
    Port = erlang:open_port({spawn, Prog}, [exit_status, in]),
    Sup ! forked,
    receive
	{Port, {exit_status, ?OTP_3906_EXIT_STATUS}} ->
	    Sup ! exited,
	    otp_3906_forker(N-1, Parent, Ref, Sup, Prog);
	{Port, Res} ->
	    exit(Res);
	Other ->
	    exit(Other)
    end.


otp_4389(suite) -> [];
otp_4389(doc) -> [];
otp_4389(Config)  when list(Config) ->
    case {os:type(),erlang:system_info(heap_type)} of
	{{unix, _},separate} ->
	    ?line Dog = test_server:timetrap(test_server:seconds(240)),
	    ?line TCR = self(),
	    case get_true_cmd() of
		True when list(True) ->
		    ?line lists:foreach(
			    fun (P) ->
				    ?line receive
					      {P, ok} ->  ?line ok;
					      {P, Err} -> ?line ?t:fail(Err)
					  end
			    end,
			    lists:map(
			      fun(_) ->
				      spawn_link(
					fun() ->
						process_flag(trap_exit, true),
						case catch open_port({spawn, True},
								     [stream,exit_status]) of
						    P when port(P) ->
							receive
							    {P,{exit_status,_}} ->
								TCR ! {self(),ok};
							    {'EXIT',_,{emfile,_}} ->
								TCR ! {self(),ok};
							    Err2 ->
								TCR ! {self(),{msg,Err2}}
							end;
						    {'EXIT',{emfile,_}} ->
							TCR ! {self(),ok};
						    Err1 ->
							TCR ! {self(), {open_port,Err1}}
						end
					end)
			      end,
			      lists:duplicate(1000,[]))),
		    ?line test_server:timetrap_cancel(Dog),
		    {comment,
		     "This test case doesn't always fail when the bug that "
		     "it tests for is present (it is most likely to fail on"
		     " a multi processor machine). If the test case fails it"
		     " will fail by deadlocking the emulator."};
		_ ->
		    ?line {skipped, "\"true\" command not found"}
	    end;
	_ ->
	    {skipped, "Only run on Unix/separate heaps"}
    end.

get_true_cmd() ->
    DoFileExist = fun (FileName) ->
 			  case file:read_file_info(FileName) of
 			      {ok, _} -> throw(FileName);
 			      _ -> not_found
 			  end
 		  end,
    catch begin
 	      %% First check in /usr/bin and /bin
 	      DoFileExist("/usr/bin/true"),
 	      DoFileExist("/bin/true"),
 	      %% Try which
 	      case filename:dirname(os:cmd("which true")) of
 		  "." -> not_found;
 		  TrueDir -> filename:join(TrueDir, "true")
 	      end
 	  end.

%% 'exit_status' option
exit_status(suite) ->
    [];
exit_status(doc) ->
    ["Test that the 'exit_status' option works"];
exit_status(Config)  when list(Config) ->
    ?line Dog = test_server:timetrap(test_server:seconds(60)),
    ?line port_expect(Config,[{"x",
			       [{exit_status, 5}]}],
		      1, "", [exit_status]),
    ?line test_server:timetrap_cancel(Dog),
    ok.

%%% Pinging functions.

stream_ping(Config, Size, CmdLine, Options) ->
    Data = random_packet(Size),
    port_expect(Config, [{Data, [Data]}], 0, CmdLine, Options).

ping(Config, Sizes, HSize, CmdLine, Options) ->
    Actions = lists:map(fun(Size) ->
				[$p|Packet] = random_packet(Size, "ping"),
				{[$p|Packet], [[$P|Packet]]}
				end,
			Sizes),
    port_expect(Config, Actions, HSize, CmdLine, Options).

%% expect_input(Sizes, HSize, CmdLine, Options)
%%
%% Sizes = Size of packets to generated.
%% HSize = Header size: 1, 2, or 4
%% CmdLine = Additional command line options.
%% Options = Addtional port options.

expect_input(Config, Sizes, HSize, CmdLine, Options) ->
    expect_input1(Config, Sizes, {HSize, CmdLine, Options}, [], []).

expect_input1(Config, [0|Rest], Params, Expect, ReplyCommand) ->
    expect_input1(Config, Rest, Params, [""|Expect], ["x0"|ReplyCommand]);
expect_input1(Config, [Size|Rest], Params, Expect, ReplyCommand) ->
    Packet = random_packet(Size),
    Fmt = io_lib:format("~c~p", [hd(Packet), Size]),
    expect_input1(Config, Rest, Params, [Packet|Expect], [Fmt|ReplyCommand]);
expect_input1(Config, [], {HSize, CmdLine0, Options}, Expect, ReplyCommand) ->
    CmdLine = build_cmd_line(CmdLine0, ReplyCommand, []),
    port_expect(Config, [{false, lists:reverse(Expect)}],
		HSize, CmdLine, Options).

build_cmd_line(FixedCmdLine, [Cmd|Rest], []) ->
    build_cmd_line(FixedCmdLine, Rest, [Cmd]);
build_cmd_line(FixedCmdLine, [Cmd|Rest], Result) ->
    build_cmd_line(FixedCmdLine, Rest, [Cmd, $:|Result]);
build_cmd_line(FixedCmdLine, [], Result) ->
    lists:flatten([FixedCmdLine, " -r", Result, " -n"]).

%% port_expect(Actions, HSize, CmdLine, Options)
%%
%% Actions = [{Send, ExpectList}|Rest]
%% HSize = 0 (stream), or 1, 2, 4   (header size aka "packet bytes")
%% CmdLine = Command line for port_test.  Don't include -h<digit>.
%% Options = Options for open_port/2.  Don't include {packet, Number} or
%%           or stream.
%%
%% Send = false | list()
%% ExpectList = List of lists or binaries.
%%
%% Returns the port.

port_expect(Config, Actions, HSize, CmdLine, Options0) ->
%    io:format("port_expect(~p, ~p, ~p, ~p)",
%		[Actions, HSize, CmdLine, Options0]),
    ?line PortTest = port_test(Config),
    ?line Cmd = lists:concat([PortTest, " -h", HSize, " ", CmdLine]),
    ?line PortType =
	case HSize of
	    0 -> stream;
	    _ -> {packet, HSize}
	end,
    ?line Options = [PortType|Options0],
    ?line io:format("open_port({spawn, ~p}, ~p)", [Cmd, Options]),
    ?line Port = open_port({spawn, Cmd}, Options),
    ?line port_expect(Port, Actions, Options),
    Port.

port_expect(Port, [{Send, Expects}|Rest], Options) when list(Expects) ->
    ?line port_send(Port, Send),
    ?line IsBinaryPort = lists:member(binary, Options),
    ?line Receiver =
	case {lists:member(stream, Options), line_option(Options)} of
	    {false, _} -> fun receive_all/2;
	    {true,false}  -> fun stream_receive_all/2;
	    {_, true} -> fun receive_all/2
	end,
    ?line Receiver(Port, maybe_to_binary(Expects, IsBinaryPort)),
    ?line port_expect(Port, Rest, Options);
port_expect(_, [], _) ->
    ok.

%%% Check for either line or {line,N} in option list
line_option([{line,X}|_]) ->
    true;
line_option([line|_]) ->
    true;
line_option([H|T]) ->
    line_option(T);
line_option([]) ->
    false.

any_list_to_binary({Atom, List}) ->
    {Atom, list_to_binary(List)};
any_list_to_binary(List) ->
    list_to_binary(List).

maybe_to_binary(Expects, true) ->
    lists:map(fun any_list_to_binary/1, Expects);
maybe_to_binary(Expects, false) ->
    Expects.

port_send(Port, false) ->
    ok;
port_send(Port, Send) when list(Send) ->
%    io:format("port_send(~p, ~p)", [Port, Send]),
    Port ! {self(), {command, Send}}.

receive_all(Port, [Expect|Rest]) ->
%    io:format("receive_all(~p, [~p|Rest])", [Port, Expect]),
    receive
	{Port, {data, Expect}} ->
	    io:format("Received ~s", [format(Expect)]),
	    ok;
	{Port, {data, Other}} ->
	    io:format("Received ~s; expected ~s",
		      [format(Other), format(Expect)]),
	    test_server:fail(bad_message);
	Other ->
	    %% (We're not yet prepared for receiving both 'eol' and
	    %% 'exit_status'; remember that they may appear in any order.)
	    case {Expect, Rest, Other} of
		{eof, [], {Port, eof}} ->
		    io:format("Received soft EOF.",[]),
		    ok;
		{{exit_status, S}, [], {Port, {exit_status, S}}} ->
		    io:format("Received exit status ~p.",[S]),
		    ok;
		_ ->
%%%	            io:format("Unexpected message: ~s", [format(Other)]),
		    io:format("Unexpected message: ~w", [Other]),
		    ?line test_server:fail(unexpected_message)
	    end
    end,
    receive_all(Port, Rest);
receive_all(_Port, []) ->
    ok.

stream_receive_all(Port, [Expect]) ->
    stream_receive_all1(Port, Expect).

stream_receive_all1(Port, Empty) when binary(Empty), size(Empty) == 0 ->
    ok;
stream_receive_all1(Port, []) ->
    ok;
stream_receive_all1(Port, Expect) ->
    receive
	{Port, {data, Data}} ->
	    Remaining = compare(Data, Expect),
	    stream_receive_all1(Port, Remaining);
	Other ->
	    test_server:fail({bad_message, Other})
    end.

compare(B1, B2) when binary(B1), binary(B2), size(B1) =< size(B2) ->
    case split_binary(B2, size(B1)) of
	{B1, Remaining} ->
	    Remaining;
	Other ->
	    test_server:fail(nomatch)
    end;
compare(B1, B2) when binary(B1), binary(B2) ->
    test_server:fail(too_much_data);
compare([X|Rest1], [X|Rest2]) ->
    compare(Rest1, Rest2);
compare([X|Rest1], [Y|Rest2]) ->
    test_server:fail(nomatch);
compare([], Remaining) ->
    Remaining;
compare(Data, []) ->
    test_server:fail(too_much_data).

maybe_to_list(Bin) when binary(Bin) ->
    binary_to_list(Bin);
maybe_to_list(List) ->
    List.

format({Eol,List}) ->
    io_lib:format("tuple<~w,~s>",[Eol, maybe_to_list(List)]);
format(List) when list(List) ->
    case list_at_least(50, List) of
	true ->
	    io_lib:format("\"~-50s...\"", [List]);
	false ->
	    io_lib:format("~p", [List])
    end;
format(Bin) when binary(Bin), size(Bin) >= 50 ->
    io_lib:format("binary<~-50s...>", [binary_to_list(Bin, 1, 50)]);
format(Bin) when binary(Bin) ->
    io_lib:format("binary<~s>", [binary_to_list(Bin)]).


list_at_least(Number, [X|Rest]) when Number > 0 ->
    list_at_least(Number-1, Rest);
list_at_least(Number, []) when Number > 0 ->
    false;
list_at_least(0, List) ->
    true.


%%% Utility functions.

random_packet(Size) ->
    random_packet(Size, "").

random_packet(Size, Prefix) ->
    build_packet(Size-length(Prefix), lists:reverse(Prefix), random_char()).

build_packet(0, Result, _NextChar) ->
    lists:reverse(Result);
build_packet(Left, Result, NextChar0) ->
    NextChar =
	if
	    NextChar0 >= 126 ->
		33;
	    true ->
		NextChar0+1
	end,
    build_packet(Left-1, [NextChar0|Result], NextChar).

sizes() ->
    case os:type() of
	vxworks ->
	    % don't stress VxWorks too much 
	    [10, 13, 64, 127, 128, 255, 256, 1023, 1024,
	     8191, 8192, 16383, 16384];
	_ ->
	    [10, 13, 64, 127, 128, 255, 256, 1023, 1024,
	     32767, 32768, 65535, 65536]
    end.

sizes(Header_Size) ->
    sizes(Header_Size, sizes(), []).

sizes(1, [Packet_Size|Rest], Result) when Packet_Size < 256 ->
    sizes(1, Rest, [Packet_Size|Result]);
sizes(2, [Packet_Size|Rest], Result) when Packet_Size < 65536 ->
    sizes(2, Rest, [Packet_Size|Result]);
sizes(4, [Packet_Size|Rest], Result) ->
    sizes(4, Rest, [Packet_Size|Result]);
sizes(_, _, Result) ->
    Result.

random_char() ->
    random_char("abcdefghijklmnopqrstuvxyzABCDEFGHIJKLMNOPQRSTUVXYZ0123456789").

random_char(Chars) ->
    lists:nth(uniform(length(Chars)), Chars).

uniform(N) ->
    case get(random_seed) of
	undefined ->
	    {X, Y, Z} = time(),
	    random:seed(X, Y, Z);
	_ ->
	    ok
    end,
    random:uniform(N).

fun_spawn(Fun) ->
    fun_spawn(Fun, []).

fun_spawn(Fun, Args) ->
    spawn_link(erlang, apply, [Fun, Args]).

port_test(Config) when list(Config) ->
    ?line filename:join(?config(data_dir, Config), "port_test").
