#include "sl_alloc_test.h"


#define CHECK_ABLK(P, MEM_SZ)						\
do {									\
    Block_t *blk_ = UMEM2BLK((P));					\
    Block_t *nxt_blk_ = NXT_BLK(blk_);					\
    long real_sz_ = ((long) nxt_blk_) - ((long) (blk_));		\
    if (IS_FREE_BLK(blk_))						\
	FAIL_TEST;							\
    if (real_sz_ > MIN_BLK_SZ						\
	&& real_sz_ > CEILING(ABLK_HDR_SZ+(MEM_SZ), sizeof(Unit_t))) {	\
	if (real_sz_							\
	    > CEILING(MIN_BLK_SZ+ABLK_HDR_SZ+(MEM_SZ), sizeof(Unit_t)))	\
	    FAIL_TEST;							\
	if (IS_NOT_LAST_BLK(blk_) && IS_FREE_BLK(nxt_blk_))		\
	    FAIL_TEST;							\
    }									\
    if (real_sz_ < CEILING(ABLK_HDR_SZ + (MEM_SZ), sizeof(Unit_t)))	\
	FAIL_TEST;							\
} while (0)


void
setup_sequence(Uint bsz, Uint no, void *res[])
{
    Block_t *blk;
    int i;

    printf("Setting up a sequence of %d blocks of size %d\n", (int) no,
	   (int) bsz);
    fflush(stdout);

    if(IS_NOT_LAST_BLK(MBC2FBLK(first_mb_carrier)))
	FAIL_TEST;

    for (i = 0; i < no; i++)
	res[i] = erts_sl_alloc(bsz);
    for (i = 0; i < no; i++)
	if (!res[i])
	    FAIL_TEST;

    printf("Checking that sequence was set up as expected\n");
    fflush(stdout);

    for (i = 1; i < no; i++)
	if (NXT_BLK(UMEM2BLK(res[i-1])) != UMEM2BLK(res[i]))
	    FAIL_TEST;

    blk = NXT_BLK(UMEM2BLK(res[no-1]));
    if (IS_NOT_LAST_BLK(blk))
	FAIL_TEST;

    printf("Sequence ok\n");
    fflush(stdout);

    /* If we fail in setup_sequence(), it doesn't mean that sl_alloc
       handles anything wrong. It is just a faulty assumption in
       setup_sequence() about how blocks are going to be placed.
       Fix setup_sequence()... */
}

void
test_free(Uint bsz)
{
    Block_t *blk;
    void *p[7];

    setup_sequence(bsz, 7, p);

    printf("Testing erts_sl_free with block size %d\n", (int) bsz);
    fflush(stdout);

    CHECK_ABLK(p[0], bsz);
    CHECK_ABLK(p[1], bsz);
    CHECK_ABLK(p[2], bsz);
    CHECK_ABLK(p[3], bsz);
    CHECK_ABLK(p[4], bsz);
    CHECK_ABLK(p[5], bsz);
    CHECK_ABLK(p[6], bsz);

    /* Coalescing with previous block */
    erts_sl_free(p[2]);
    erts_sl_free(p[3]); 

    blk = NXT_BLK(UMEM2BLK(p[1]));
    if(!IS_FREE_BLK(blk))
	FAIL_TEST;
    if (NXT_BLK(blk) != UMEM2BLK(p[4]))
	FAIL_TEST;

    /* Coalescing with next block */

    erts_sl_free(p[1]);
    blk = NXT_BLK(UMEM2BLK(p[0]));
    if(!IS_FREE_BLK(blk))
	FAIL_TEST;
    if (NXT_BLK(blk) != UMEM2BLK(p[4]))
	FAIL_TEST;
    
    /* Coalescing with next and previous block */

    erts_sl_free(p[5]);
    erts_sl_free(p[4]);

    blk = NXT_BLK(UMEM2BLK(p[0]));
    if(!IS_FREE_BLK(blk))
	FAIL_TEST;
    if (NXT_BLK(blk) != UMEM2BLK(p[6]))
	FAIL_TEST;
    
    /* Cleanup */

    erts_sl_free(p[0]);
    erts_sl_free(p[6]);

    printf("erts_sl_free with block size %d succeded\n", (int) bsz);
    fflush(stdout);

}

void
test_realloc(Uint bsz)
{
    Block_t *blk;
    Block_t *tmp_blk;
    void *p[3];
    Uint tmp_sz;

    setup_sequence(bsz, 3, p);

    CHECK_ABLK(p[0], bsz);
    CHECK_ABLK(p[1], bsz);
    CHECK_ABLK(p[2], bsz);

    printf("Testing erts_sl_realloc with block size %d\n", (int) bsz);
    fflush(stdout);

    erts_sl_free(p[1]);

    blk = NXT_BLK(UMEM2BLK(p[0]));
    if (!IS_FREE_BLK(blk))
	FAIL_TEST;

    /* Shrink and coalecse with next free */

    if (p[0] != erts_sl_realloc(p[0], bsz, bsz/2))
	FAIL_TEST;

    CHECK_ABLK(p[0], bsz/2);

    blk = NXT_BLK(UMEM2BLK(p[0]));
    if (!IS_FREE_BLK(blk))
	FAIL_TEST;
    if (NXT_BLK(blk) != UMEM2BLK(p[2]))
	FAIL_TEST;

    /* Grow into next free; but leave free block at end */

    if (p[0] != erts_sl_realloc(p[0], bsz/2, bsz*3/4))
	FAIL_TEST;

    CHECK_ABLK(p[0], bsz*3/4);
    blk = NXT_BLK(UMEM2BLK(p[0]));

    if (!IS_FREE_BLK(blk))
	FAIL_TEST;
    if (NXT_BLK(blk) != UMEM2BLK(p[2]))
	FAIL_TEST;

    /* Grow upto next alloced block by allocating just enough so that no
       free block fits between them */
    tmp_sz = BLK_SZ(blk);
    blk = UMEM2BLK(p[0]);
    tmp_sz += BLK_SZ(blk) - ABLK_HDR_SZ;

    if (p[0] != erts_sl_realloc(p[0], bsz*3/4, tmp_sz-(MIN_BLK_SZ-1)))
	FAIL_TEST;

    CHECK_ABLK(p[0], tmp_sz-(MIN_BLK_SZ-1));
    blk = NXT_BLK(UMEM2BLK(p[0]));

    if (IS_FREE_BLK(blk))
	FAIL_TEST;
    if (blk != UMEM2BLK(p[2]))
	FAIL_TEST;

    /* Grow into unused part at end */
    if (p[0] != erts_sl_realloc(p[0], tmp_sz-(MIN_BLK_SZ-1), tmp_sz))
	FAIL_TEST;

    CHECK_ABLK(p[0], tmp_sz);
    blk = NXT_BLK(UMEM2BLK(p[0]));
    if (IS_FREE_BLK(blk))
	FAIL_TEST;
    if (blk != UMEM2BLK(p[2]))
	FAIL_TEST;

    /* Shrink *almost* as much so that a free block would fit between the
       allocated blocks, and make sure that we don't get a free block
       in between */
    if (p[0] != erts_sl_realloc(p[0], tmp_sz, tmp_sz-(MIN_BLK_SZ-1)))
	FAIL_TEST;

    CHECK_ABLK(p[0], tmp_sz-(MIN_BLK_SZ-1));
    blk = NXT_BLK(UMEM2BLK(p[0]));

    if (IS_FREE_BLK(blk))
	FAIL_TEST;
    if (blk != UMEM2BLK(p[2]))
	FAIL_TEST;

    /* Shrink just as much so that a free block can fit between
       the alloced blocks */
    if (p[0] != erts_sl_realloc(p[0],
				tmp_sz-(MIN_BLK_SZ-1),
				tmp_sz-MIN_BLK_SZ))
	FAIL_TEST;

    CHECK_ABLK(p[0], tmp_sz-MIN_BLK_SZ);
    blk = NXT_BLK(UMEM2BLK(p[0]));

    if (IS_ALLOCED_BLK(blk))
	FAIL_TEST;
    if (NXT_BLK(blk) != UMEM2BLK(p[2]))
	FAIL_TEST;

    /* Shrink so little that no free block would fit between allocated
       blocks, and make sure that we shrink the allocated block and
       coalesce the extra free part with the next free block. */
    tmp_blk = blk;
    if (p[0] != erts_sl_realloc(p[0],
				tmp_sz-MIN_BLK_SZ,
				tmp_sz-(2*MIN_BLK_SZ-1)))
	FAIL_TEST;

    CHECK_ABLK(p[0], tmp_sz-(2*MIN_BLK_SZ-1));
    blk = NXT_BLK(UMEM2BLK(p[0]));

    if (IS_ALLOCED_BLK(blk))
	FAIL_TEST;
    if (tmp_blk == blk)
	FAIL_TEST;
    if (NXT_BLK(blk) != UMEM2BLK(p[2]))
	FAIL_TEST;

    /* Cleanup */
    erts_sl_free(p[0]);
    erts_sl_free(p[2]);

    printf("erts_sl_realloc with block size %d succeded\n", (int) bsz);
    fflush(stdout);

}

START_TEST
{
    int max;
    int i;
    ErtsSlAllocInit esai;

    if (!test_new_sl_alloc)
	SKIP_TEST("Cannot test old sl_alloc");

    memset((void *) &esai, 255, sizeof(ErtsSlAllocInit));
    esai.esla = 1;
    esai.eosla = 0;
    esai.sbct = 100000000;

    erts_sl_alloc_init(&esai);

    max = esai.mcs/7-sizeof(MBCarrier_t)-7*ABLK_HDR_SZ*sizeof(Unit_t);

    for (i = 100; i <= max; i += 499)
	test_free(i);

    for (i = 100; i <= max; i += 499)
	test_realloc(i);

    TEST_SUCCEEDED;
}
END_TEST
