%% ``The contents of this file are subject to the Erlang Public License,
%% Version 1.1, (the "License"); you may not use this file except in
%% compliance with the License. You should have received a copy of the
%% Erlang Public License along with this software. If not, it can be
%% retrieved via the world wide web at http://www.erlang.org/.
%% 
%% Software distributed under the License is distributed on an "AS IS"
%% basis, WITHOUT WARRANTY OF ANY KIND, either express or implied. See
%% the License for the specific language governing rights and limitations
%% under the License.
%% 
%% The Initial Developer of the Original Code is Ericsson Utvecklings AB.
%% Portions created by Ericsson are Copyright 1999, Ericsson Utvecklings
%% AB. All Rights Reserved.''
%% 
%%     $Id$
%%
-module(vector_SUITE).

-export([all/1,init_per_testcase/2,fin_per_testcase/2,
	 old/1,eq/1,lt/1,ge/1,circular/1]).

-include("test_server.hrl").

-define(DEFAULT_TIMEOUT, ?t:minutes(2)).

all(suite) -> [old,eq,lt,ge,circular].

init_per_testcase(Case, Config) ->
    ?line Dog = test_server:timetrap(?DEFAULT_TIMEOUT),
    [{watchdog, Dog}|Config].

fin_per_testcase(Case, Config) ->
    Dog=?config(watchdog, Config),
    test_server:timetrap_cancel(Dog),
    ok.

old(Config) when list(Config) ->
    Sizes = lists:seq(1, 300),
    old_1(Sizes, vector:new(1, {})).

old_1([Sz|Sizes], Prev) ->
    io:format("Size: ~p\n", [Sz]),
    roundtrip(Prev),
    PrevList = vector:to_list(Prev),
    T0 = random_term(),
    V0List = lists:duplicate(Sz, T0),
    V0 = vector:new(Sz, T0),
    V0List = vector:to_list(V0),
    Pos = random:uniform(size(V0)),
    V1 = vector:set(Pos, V0, Prev),
    V0List = vector:to_list(V0),
    _ = vector:set(1, Prev, 455.35),
    _ = vector:set(1, Prev, now()),
    _ = vector:set(1, Prev, Prev),
    PrevList = vector:to_list(Prev),
    roundtrip(V0),
    V0 = vector:from_list(V0List),
    old_1(Sizes, V1);
old_1([], _) -> ok.

eq(Config) when list(Config) ->
    ?line V = vector:from_list([a,b,c]),
    ?line eq(true, V, V),
    ?line eq(false, V, 42),
    ?line eq(false, -1, V),
    ?line eq(false, dum, V),

    ?line eq(false, [a,b,c], [x,b,c]),
    ?line eq(false, [a,b,c], [a,x,c]),
    ?line eq(false, [a,b,c], [a,b,x]),

    ?line eq(true, [], []),
    ?line eq(false, [a], []),
    ?line eq(false, [], [a]),

    ?line V2 = vector:set(1, V, V),
    ?line V3 = vector:from_list([V,b,c]),
    ?line V2 = V3,
    
    ok.

eq(R, V1, V2) when list(V1) ->
    eq(R, vector:from_list(V1), V2);
eq(R, V1, V2) when list(V2) ->
    eq(R, V1, vector:from_list(V2));
eq(R, V1, V2) ->
    R = V1 =:= V2,
    R = V1 == V2.

lt(Config) when list(Config) ->
    ?line lt(false, [], []),
    ?line lt(true, [], [1.0]),
    ?line lt(true, [1.0], [2]),
    ?line lt(true, [0,1.0], [0.0,2]),
    ?line lt(true, [0,a,1], [0.0,a,2.0]),
    ?line lt(true, {a,b}, [a,b]),
    ok.

lt(R, V1, V2) when list(V1) ->
    lt(R, vector:from_list(V1), V2);
lt(R, V1, V2) when list(V2) ->
    lt(R, V1, vector:from_list(V2));
lt(R, V1, V2) ->
    R = V1 < V2.
 
ge(Config) when list(Config) ->
    ?line ge(true, [], []),
    ?line ge(false, [], [42]),
    ?line ge(false, [1.0], [2]),
    ?line ge(true, [0,1.0], [0.0,1]),
    ?line ge(false, [0,a,1], [0.0,a,2.0]),
    ?line ge(false, {a,b}, [a,b]),
    ok.

ge(R, V1, V2) when list(V1) ->
    ge(R, vector:from_list(V1), V2);
ge(R, V1, V2) when list(V2) ->
    ge(R, V1, vector:from_list(V2));
ge(R, V1, V2) ->
    R = V1 >= V2.

circular(Config) when list(Config) ->
    DummyFun = fun() -> Config end,
    ?line V = vector:from_list([a,b,DummyFun]),
    ?line roundtrip(V),
    ?line V2 = vector:set(1, V, V),
    ?line [V,b,DummyFun] = vector:to_list(V2),
    ?line roundtrip(V2),
    ?line V3 = vector:set(2, V2, V2),
    ?line roundtrip(V3),
    ?line roundtrip(V2),
    ?line roundtrip(V),
    ?line roundtrips(V3, 150, []),
    ?line roundtrips(V2, 150, []),
    ?line roundtrips2(V2, 300, []),
    ?line roundtrips2(V2, 300, []),
    ok.

roundtrips(V, 0, Acc) -> ok;
roundtrips(V, I, Acc) ->
    roundtrip(vector:from_list(Acc)),
    roundtrips(V, I-1, [V|Acc]).

roundtrips2(V, 0, Acc) -> ok;
roundtrips2(V, I, Acc) ->
    roundtrip(vector:from_list([V,list_to_tuple(Acc)])),
    roundtrips2(V, I-1, [I|Acc]).

roundtrip(Data) ->
    Self = self(),
    Pid = spawn_link(fun() ->
			     receive
				 {data,Data} -> Self ! Data
			     end
		     end),
    Pid ! {data,Data},
    receive
	Data -> ok;
	Other -> ?t:fail({expected,Data,got,Other})
    end.

%% Create an arbitrary element/term.
random_term() ->
    F = random:uniform(),
    case random:uniform(13) of
	1 -> random:uniform(16#ffffff);
	2 -> random:uniform(16#ffffff)* 9999999999999999999999;
	3 -> [random:uniform(),1];
	4 -> make_ref();
	5 -> vector:new(random:uniform(10), random:uniform());
	6 -> list_to_binary([1,2,3,4,random:uniform(255)]);
	7 -> random:uniform();
	8 -> {a,random:uniform(332423),random:uniform()};
	9 -> list_to_atom("an_random_atom" ++ integer_to_list(random:uniform(50)));
	10 -> self();
	11 -> fun(X) -> 2*X+F end;
	12 -> 438.3748;
	13 -> 392879248729842948729847298472974297
    end.
